from typing import Callable, Dict, Union, List, Tuple, get_args, Any, Sequence

import numpy as np
from approvaltests.scrubbers.scrubbers import Scrubber, create_regex_scrubber, combine_scrubbers, scrub_all_dates, \
    scrub_all_guids

JsonLikeCollection = Union[Dict, List, Tuple]
RecursiveScrubber = Callable[[JsonLikeCollection], JsonLikeCollection]
SequenceScrubber = Callable[[Sequence], Sequence]


def scrub_recursive(tags: JsonLikeCollection, scrubber: Scrubber) -> JsonLikeCollection:
    if isinstance(tags, Dict):
        return scrub_dict_recursive(tags, scrubber)
    if isinstance(tags, List):
        return scrub_list_recursive(tags, scrubber)
    if isinstance(tags, Tuple):
        return scrub_tuple_recursive(tags, scrubber)


def scrub_dict_recursive(tags: Dict, scrubber: Scrubber) -> Dict:
    scrubbed = {}
    for k, v in tags.items():
        if isinstance(k, str):
            k = scrubber(k)
        v = scrub_element(v, scrubber)
        scrubbed[k] = v
    return scrubbed


def scrub_element(element: Any, scrubber: Scrubber) -> Any:
    if isinstance(element, str):
        element = scrubber(element)
    if isinstance(element, get_args(JsonLikeCollection)):
        element = scrub_recursive(element, scrubber)
    return element


def scrub_list_recursive(tags: List, scrubber: Scrubber) -> List:
    return [scrub_element(t, scrubber) for t in tags]


def scrub_tuple_recursive(tags: Tuple, scrubber: Scrubber) -> Tuple:
    return tuple(scrub_element(t, scrubber) for t in tags)


def scrub_sequential(elems: Sequence, scrubber: Scrubber) -> Sequence:
    return [scrubber(e) for e in elems]


def make_scrubber_recurse(scrubber: Scrubber) -> RecursiveScrubber:
    return lambda elems: scrub_recursive(elems, scrubber)


def make_scrubber_sequential(scrubber: Scrubber) -> SequenceScrubber:
    return lambda elems: scrub_sequential(elems, scrubber)


def identity_recursive_scrubber(tags: Dict) -> Dict:
    return tags


def identity_sequence_scrubber(v: Sequence) -> Sequence:
    return v


def scrub_xarray_metadata(a, tags_scrubber):
    a.attrs = tags_scrubber(a.attrs)
    for name in a.coords:
        a[name].attrs = tags_scrubber(a[name].attrs)
    for name in getattr(a, 'data_vars', {}):
        a[name].attrs = tags_scrubber(a[name].attrs)

    return a


def scrub_xarray_coordinates(a, coords_scrubber):
    for coord in a.coords:
        cv = a[coord]
        if cv.dtype.type is np.str_:
            a = a.assign_coords({coord: (cv.dims[0], coords_scrubber(cv.values))})
    return a


def scrub_all_yeoda_dates(data: str) -> str:
    return create_regex_scrubber(
        r"\d{4}\d{2}\d{2}T\d{2}\d{2}\d{2}",
        lambda t: f"<yeoda_date{t}>",
    )(data)


def scrub_all_short_commits(data: str) -> str:
    return create_regex_scrubber(
        r"[0-9a-fA-F]{7}",
        lambda t: f"<short_commit_{t}>",
    )(data)


def scrub_all_tags(data: str) -> str:
    return create_regex_scrubber(
        r"v\d\.\d\.\d",
        lambda t: f"<tag_{t}>",
    )(data)


def scrub_yeoda_datacube_metadata(data: str) -> str:
    return combine_scrubbers(
        scrub_all_dates,
        scrub_all_yeoda_dates,
        scrub_all_guids,
        scrub_all_short_commits,
        scrub_all_tags,
    )(data)



from dataclasses import dataclass
from typing import Mapping


@dataclass
class Tolerance:
    rel: float = 1e-09
    abs: float = 0.0

    def to_kwargs(self) -> Mapping:
        return dict(rtol=self.rel, atol=self.abs)



from datetime import datetime

import numpy as np
import pytest
from approvaltests.scrubbers import create_regex_scrubber

from factories import make_zarr_at
from pytest_approvaltests_geo.reporters.report_geo_zarrs import ReportGeoZarrs
from pytest_approvaltests_geo.scrubbers import make_scrubber_recurse, make_scrubber_sequential


@pytest.fixture
def reporter():
    return ReportGeoZarrs()


def test_identical_geo_zarrs_report_nothing(reporter, tmp_path, capsys):
    geo_zarr = make_zarr_at([[42]], tmp_path / "geo_zarr.zarr", dict(some='tag'))
    assert reporter.report(geo_zarr.as_posix(), geo_zarr.as_posix())
    assert capsys.readouterr().out == ""


def test_report_pixel_differences_and_statistics(reporter, tmp_path, capsys):
    received = make_zarr_at([[0, 0], [0, np.nan]], tmp_path / "received.zarr", dict(some='tag'))
    approved = make_zarr_at([[2, 6], [-1, 2]], tmp_path / "approved.zarr", dict(some='tag'))
    assert reporter.report(received.as_posix(), approved.as_posix())
    std_out = capsys.readouterr().out
    assert "min=1.0, max=6.0, mean=3.0, median=2.0, nans=-1" in std_out


def test_report_meta_data_differences(reporter, tmp_path, capsys):
    received = make_zarr_at([[42]], tmp_path / "received.zarr", dict(some='tag'))
    approved = make_zarr_at([[42]], tmp_path / "approved.zarr", dict(some='other'))
    assert reporter.report(received.as_posix(), approved.as_posix())
    output = capsys.readouterr().out
    assert 'L   some: tag' in output and 'R   some: other' in output


def test_report_scrubbed_data(tmp_path, capsys):
    date_scrubber = create_regex_scrubber(
        r"\d{4}-\d{2}-\d{2}T\d{2}-\d{2}-\d{2}", lambda t: f"<date{t}>"
    )
    scrubbing_reporter = ReportGeoZarrs(make_scrubber_recurse(date_scrubber), make_scrubber_sequential(date_scrubber))
    date_old = datetime(2022, 1, 1).strftime("%Y-%m-%dT%H-%M-%S")
    date_new = datetime(2022, 2, 1).strftime("%Y-%m-%dT%H-%M-%S")
    received = make_zarr_at([[0, 0], [0, 0]], tmp_path / "received.zarr",
                            {"some": date_old}, {date_new: 42}, extra_coords={'meta': ('band', [date_old])})
    approved = make_zarr_at([[2, 4], [-2, 1]], tmp_path / "approved.zarr",
                            {"other": date_old}, {date_new: 21}, extra_coords={'meta': ('band', [date_new])})
    assert scrubbing_reporter.report(received.as_posix(), approved.as_posix())
    output = capsys.readouterr().out
    assert 'other: <date0>' in output and '<date0>: 42' in output
    assert 'some: <date0>' in output and '<date0>: 21' in output
    assert "(band) <U19 '2022-01-01T00-00-00'" not in output



from datetime import datetime

import pytest
from approvaltests.scrubbers import create_regex_scrubber
from xarray import DataArray

from factories import make_zarr_at
from pytest_approvaltests_geo.comparators.compare_geo_zarrs import CompareGeoZarrs
from pytest_approvaltests_geo.float_utils import Tolerance
from pytest_approvaltests_geo.scrubbers import make_scrubber_recurse, make_scrubber_sequential


@pytest.fixture
def comparator():
    return CompareGeoZarrs()


def test_compare_identical_geo_zarrs(comparator, tmp_path):
    geo_zarr = make_zarr_at([[42]], tmp_path / "geo_zarr.zarr", dict(some='tag'))
    assert comparator.compare(geo_zarr.as_posix(), geo_zarr.as_posix())


def test_compare_geo_zarrs_with_differing_attrs(comparator, tmp_path):
    received = make_zarr_at([[42]], tmp_path / "received.zarr", dict(some='tag'))
    approved = make_zarr_at([[42]], tmp_path / "approved.zarr", dict(some='other'))
    assert not comparator.compare(received.as_posix(), approved.as_posix())


def test_compare_geo_zarrs_with_differing_data_var_attrs(comparator, tmp_path):
    received = make_zarr_at([[42]], tmp_path / "received.zarr", array_attrs=dict(some='tag'))
    approved = make_zarr_at([[42]], tmp_path / "approved.zarr", array_attrs=dict(some='other'))
    assert not comparator.compare(received.as_posix(), approved.as_posix())


def test_compare_geo_zarrs_with_differing_coord_attrs(comparator, tmp_path):
    received = make_zarr_at([[42]], tmp_path / "received.zarr", coords=dict(
        band=[1], x=[0], y=[0], spatial_ref=DataArray(0, attrs=dict(some='tag'))
    ))
    approved = make_zarr_at([[42]], tmp_path / "approved.zarr", coords=dict(
        band=[1], x=[0], y=[0], spatial_ref=DataArray(0, attrs=dict(some='other'))
    ))
    assert not comparator.compare(received.as_posix(), approved.as_posix())


def test_compare_geo_zarrs_with_differing_pixels(comparator, tmp_path):
    received = make_zarr_at([[42]], tmp_path / "received.zarr", dict(some='tag'))
    approved = make_zarr_at([[21]], tmp_path / "approved.zarr", dict(some='tag'))
    assert not comparator.compare(received.as_posix(), approved.as_posix())


def test_compare_geo_zarrs_applies_scrubbers_to_all_meta_data(tmp_path):
    date_scrubber = create_regex_scrubber(r"\d{4}-\d{2}-\d{2}T\d{2}-\d{2}-\d{2}", lambda t: f"<date{t}>")
    scrubbing_comparator = CompareGeoZarrs(make_scrubber_recurse(date_scrubber))
    date_old = datetime(2022, 1, 1).strftime("%Y-%m-%dT%H-%M-%S")
    date_new = datetime(2022, 1, 2).strftime("%Y-%m-%dT%H-%M-%S")
    received = make_zarr_at([[42]], tmp_path / "received.zarr", {date_old: 42}, dict(some=date_old), coords=dict(
        band=[1], x=[0], y=[0], spatial_ref=DataArray(0, attrs=dict(some=date_old))
    ))
    approved = make_zarr_at([[42]], tmp_path / "approved.zarr", {date_new: 42}, dict(some=date_new), coords=dict(
        band=[1], x=[0], y=[0], spatial_ref=DataArray(0, attrs=dict(some=date_new))
    ))
    assert scrubbing_comparator.compare(received.as_posix(), approved.as_posix())

def test_compare_geo_zarrs_applies_scrubbers_to_all_coords(tmp_path):
    date_scrubber = create_regex_scrubber(r"\d{4}-\d{2}-\d{2}T\d{2}-\d{2}-\d{2}", lambda t: f"<date{t}>")
    scrubbing_comparator = CompareGeoZarrs(coords_scrubber=make_scrubber_sequential(date_scrubber))
    date_old = datetime(2022, 1, 1).strftime("%Y-%m-%dT%H-%M-%S")
    date_new = datetime(2022, 1, 2).strftime("%Y-%m-%dT%H-%M-%S")
    received = make_zarr_at([[42]], tmp_path / "received.zarr", extra_coords={'meta': ('band', [date_old])})
    approved = make_zarr_at([[42]], tmp_path / "approved.zarr", extra_coords={'meta': ('band', [date_new])})
    assert scrubbing_comparator.compare(received.as_posix(), approved.as_posix())


def test_compare_geo_zarrs_floats_with_tolerances(tmp_path):
    tolerant_comparator = CompareGeoZarrs(float_tolerance=Tolerance(rel=0.008, abs=0.0021))
    received = make_zarr_at([[-1.01]], tmp_path / "received.zarr")
    approved = make_zarr_at([[-1.0]], tmp_path / "approved.zarr")
    assert tolerant_comparator.compare(received.as_posix(), approved.as_posix())



from pathlib import Path
from typing import Optional, Union

import pytest
import rasterio
from approval_utilities.utilities.exceptions.exception_collector import gather_all_exceptions_and_throw
from approvaltests import verify_with_namer_and_writer, ExistingFileWriter, ScenarioNamer
from approvaltests.namer import NamerBase
from pytest_approvaltests_geo._version import __version__
from pytest_approvaltests_geo.comparators.compare_geo_ncs import CompareGeoNcs
from pytest_approvaltests_geo.comparators.compare_geo_tiffs import CompareGeoTiffs
from pytest_approvaltests_geo.comparators.compare_geo_zarrs import CompareGeoZarrs
from pytest_approvaltests_geo.existing_dir_writer import ExistingDirWriter
from pytest_approvaltests_geo.geo_options import GeoOptions
from pytest_approvaltests_geo.namer.stack_frame_namer_with_external_data_dir import StackFrameNamerWithExternalDataDir
from pytest_approvaltests_geo.reporters.report_geo_ncs import ReportGeoNcs
from pytest_approvaltests_geo.reporters.report_geo_tiffs import ReportGeoTiffs
from pytest_approvaltests_geo.reporters.report_geo_zarrs import ReportGeoZarrs
from pytest_approvaltests_geo.scrubbers import RecursiveScrubber
from xarray import DataArray

APPROVAL_TEST_GEO_DATA_ROOT_OPTION = "--approval-test-geo-data-root"

PathConvertible = Union[Path, str]


def pytest_addoption(parser):
    group = parser.getgroup('pytest-approvaltests-geo')
    group.addoption(APPROVAL_TEST_GEO_DATA_ROOT_OPTION,
                    default=None,
                    help="specify local approval test data root")

    parser.addini('approvaltests_geo_data_root',
                  'Path to your approval test geo data root containing your input and approved files', type='string')
    parser.addini('approvaltests_geo_input',
                  'Subdirectory within the approval test geo data root containing your input files', type='string')
    parser.addini('approvaltests_geo_approved',
                  'Subdirectory within the approval test geo data root containing your approved files', type='string')


def get_approval_root(config):
    custom_root = config.option.approval_test_geo_data_root
    if custom_root is not None:
        return Path(custom_root)

    root = config.getini('approvaltests_geo_data_root')
    if root:
        return Path(root)
    return None


def pytest_collection_modifyitems(config, items):
    approval_root = get_approval_root(config)
    if approval_root:
        skip_reason = f"needs {approval_root} to run."
    else:
        skip_reason = "needs an approval directory configured to run."
    skip_missing_dataset = pytest.mark.skip(reason=skip_reason)
    for item in items:
        if "approval_test_geo_data_root" in getattr(item, 'fixturenames', {}) and \
                (approval_root is None or not approval_root.exists()):
            item.add_marker(skip_missing_dataset)


@pytest.fixture(scope='module')
def approval_test_geo_data_root(request):
    custom_root = request.config.option.approval_test_geo_data_root
    if custom_root is not None:
        return Path(custom_root)

    root = request.config.getini('approvaltests_geo_data_root')
    if root:
        return Path(root)
    return None


@pytest.fixture(scope='module')
def approval_geo_input_directory(approval_test_geo_data_root, request):
    if approval_test_geo_data_root is not None:
        return approval_test_geo_data_root / request.config.getini('approvaltests_geo_input')
    return None


@pytest.fixture(scope='module')
def approved_geo_directory(approval_test_geo_data_root, request):
    if approval_test_geo_data_root is not None:
        return approval_test_geo_data_root / request.config.getini('approvaltests_geo_approved')
    return None


@pytest.fixture(scope='module')
def geo_data_namer_factory(approved_geo_directory):
    if approved_geo_directory is not None:
        return lambda: StackFrameNamerWithExternalDataDir(approved_geo_directory.as_posix())
    else:
        from approvaltests.namer.default_name import get_default_namer
        return get_default_namer


@pytest.fixture(scope='module')
def name_geo_scenario(geo_data_namer_factory):
    def scenario_namer(*scenario_names):
        return ScenarioNamer(geo_data_namer_factory(), *scenario_names)

    return scenario_namer


@pytest.fixture(scope='module')
def verify_geo_tif(verify_geo_tif_with_namer, geo_data_namer_factory):
    def _verify_fn(tile_file: PathConvertible,
                   *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
                   options: Optional[GeoOptions] = None):
        geo_data_namer = options.namer if options else None
        geo_data_namer = geo_data_namer or geo_data_namer_factory()
        geo_data_namer.set_extension(Path(tile_file).suffix)
        verify_geo_tif_with_namer(tile_file, geo_data_namer, options=options)

    return _verify_fn


@pytest.fixture(scope='module')
def verify_geo_tif_with_namer():
    def _verify_fn(tile_file: PathConvertible,
                   namer: NamerBase,
                   *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
                   options: Optional[GeoOptions] = None):
        options = options or GeoOptions()
        tif_comparator = CompareGeoTiffs(options.scrub_tags, options.tolerance)
        tif_reporter = ReportGeoTiffs(options.scrub_tags, options.tolerance)
        if options.has_scenario_by_tags():
            with rasterio.open(tile_file) as rds:
                namer = options.wrap_namer_in_tags_scenario(namer, rds.tags())
        options = options.with_comparator(tif_comparator)
        options = options.with_reporter(tif_reporter)
        verify_with_namer_and_writer(
            namer=namer,
            writer=ExistingFileWriter(tile_file, options),
            options=options)

    return _verify_fn


@pytest.fixture(scope='module')
def verify_raster_as_geo_tif(verify_geo_tif, tmp_path_factory):
    def _verify_fn(tile: DataArray,
                   *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
                   options: Optional[GeoOptions] = None):
        tile_file = tmp_path_factory.mktemp("raster_as_geo_tif") / "raster.tif"
        options = options or GeoOptions()
        options.tif_writer(tile_file, tile)
        verify_geo_tif(tile_file, options=options)

    return _verify_fn


@pytest.fixture(scope='module')
def verify_geo_zarr(geo_data_namer_factory):
    def _verify_fn(zarr_archive: PathConvertible,
                   *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
                   options: Optional[GeoOptions] = None):
        options = options or GeoOptions()
        geo_data_namer = options.namer or geo_data_namer_factory()
        geo_data_namer.set_extension(Path(zarr_archive).suffix)
        zarr_comparator = CompareGeoZarrs(options.scrub_tags, options.scrub_coords, options.tolerance)
        zarr_reporter = ReportGeoZarrs(options.scrub_tags, options.scrub_coords, options.tolerance)
        options = options.with_comparator(zarr_comparator)
        options = options.with_reporter(zarr_reporter)
        verify_with_namer_and_writer(
            namer=geo_data_namer,
            writer=ExistingDirWriter(zarr_archive),
            options=options)

    return _verify_fn


@pytest.fixture(scope='module')
def verify_geo_nc(geo_data_namer_factory):
    def _verify_fn(nc_file: PathConvertible,
                   *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
                   options: Optional[GeoOptions] = None):
        options = options or GeoOptions()
        geo_data_namer = options.namer or geo_data_namer_factory()
        geo_data_namer.set_extension(Path(nc_file).suffix)
        nc_comparator = CompareGeoNcs(options.scrub_tags, options.scrub_coords, options.tolerance)
        nc_reporter = ReportGeoNcs(options.scrub_tags, options.scrub_coords, options.tolerance)
        options = options.with_comparator(nc_comparator)
        options = options.with_reporter(nc_reporter)
        verify_with_namer_and_writer(
            namer=geo_data_namer,
            writer=ExistingFileWriter(nc_file, options),
            options=options)

    return _verify_fn


@pytest.fixture
def verify_data_frame_using(verify_geo_tif, geo_data_namer_factory):
    def _make_verifier(verify_fn, *columns):
        def _verify_data_frame(data_frame, *, options: Optional[GeoOptions] = None):
            options = options or GeoOptions()

            def _verify(pack):
                _, row = pack
                verify_fn(row['filepath'], options=GeoOptions.from_options(options.with_namer(
                    ScenarioNamer(geo_data_namer_factory(), *tuple(row[c] for c in columns))
                )))

            gather_all_exceptions_and_throw(data_frame.iterrows(), _verify)

        return _verify_data_frame

    return _make_verifier



import re
from pathlib import Path

from setuptools import setup

version_file = Path(__file__).parent / "src/pytest_approvaltests_geo/_version.py"
version_line = re.search(r'__version__ = ["\']\d\.\d\.\d(\.dev\d)?["\']', version_file.read_text()).group(0)
version_str = re.search(r'\d\.\d\.\d(\.dev\d)?', version_line).group(0)
setup(version=version_str)